' =============================================================================
' TITLE:	MSR-H01 Hexapod & HexEngine Binary object avoidance
' AUTHOR:	James Dickinson
' DATE:		28/06/2011
' REVISION:	3.0
' 
' =============================================================================

' CONSTANTS ===================================================================
'
' ANY CONSTANTS USED FOR HEAD, BODY OR WALK CAN HAVE A MAX OR MIN VALUE OF -127/127

cMAX_WALK_SPEED	= 80	' MAXIMUM OF 127
cMAX_CRAB_SPEED = 40    ' MAXIMUM OF 127
cMAX_TURN_SPEED	= 40	' MAXIMUM OF 127

cMAX_HEAD_PAN = 127
cMIN_HEAD_PAN = -127
cMAX_HEAD_TILT = 100
cMIN_HEAD_TILT = -127

' BODY LIMITS MAXIMUM OF +/- 127
cMAX_BODY_Y_TRANSLATE = 100
cMIN_BODY_Y_TRANSLATE = -100
cMAX_BODY_Z_TRANSLATE = 50
cMIN_BODY_Z_TRANSLATE = -50
cMAX_BODY_X_ROTATE = 100
cMIN_BODY_X_ROTATE = -70
cMAX_BODY_Z_ROTATE = 100
cMIN_BODY_Z_ROTATE = -100

PIPPan = "~"
PIPBody = "~"


' VARIABLES ===================================================================

' get the current center of the image
midx = GetVariable("IMAGE_WIDTH") / 2
midy = GetVariable("IMAGE_HEIGHT") / 2

' PIP 'H' COMMAND VARIABLES
VariableH1 = 1
VariableH2 = 1

' PIP 'B' COMMAND VARIABLES
VariableB1 = 1
VariableB2 = 1
VariableB3 = 1
VariableB4 = 1
VariableB5 = 1
VariableB6 = 1

' PIP 'M' COMMAND VARIABLES
VariableM1 = 1
VariableM2 = 1
VariableM3 = 1
	
' BODY ROTATE & TRANSLATE
gBodyXRotate = 0
gBodyZRotate = 0
gBodyYTranslate = 0

gPowerState = GetVariable("POWER_STATE")

'=======BINARY OBJECT AVOIDANCE START

width = GetVariable("IMAGE_WIDTH")
height = GetVariable("IMAGE_HEIGHT")

'Grab information for Harris Corners
harris_corners = GetArrayVariable("HARRIS_CORNERS")
harris_size = UBound(harris_corners)

'initialise point arrays. maximum of 200 points (adjust this or harris corners [to reduce point count] for max_index errors)
if (VariableExists("STARTUP") = false) then
  dim pointx(200)
  dim pointy(200)
  setvariable "STARTUP", 1
end if

'pull point data from harris corners
max_index = 0
for index = 0 to harris_size
  pointx(max_index) = harris_corners(index)
  index = index + 1
  pointy(max_index) = harris_corners(index)
  max_index = max_index + 1
next
max_index = max_index - 1

' reset point variables
furthest_x = 0
furthest_y = 0
left_heighest_x = 0
left_heighest_y = 0
right_heighest_x = 0
right_heighest_y = 0
dangerx = 9999
dangery = 9999
left_danger_x = 9999
left_danger_y = 9999
right_danger_x = 0
right_danger_y = 9999

' find points in array
for index = 0 to max_index
  if pointy(index) > furthest_y then
    furthest_y = pointy(index)
    furthest_x = pointx(index)
  end if
  if pointx(index) < (width / 2) then
    if pointy(index) > left_heighest_y then
      left_heighest_y = pointy(index)
      left_heighest_x = pointx(index)
    end if
  else
    if pointy(index) > right_heighest_y then
      right_heighest_y = pointy(index)
      right_heighest_x = pointx(index)
    end if
  end if
  if pointy(index) < dangery then
    dangery = pointy(index)
    dangerx = pointx(index)
  end if
  if pointx(index) <= (width * 0.3) then
    if pointy(index) < left_danger_y then
      left_danger_x = pointx(index)
      left_danger_y = pointy(index)
    end if
  end if
  if pointx(index) >= (width * 0.7) then
    if pointy(index) < right_danger_y then
      right_danger_x = pointx(index)
      right_danger_y = pointy(index)
    end if
  end if
next

'how close the danger point needs to be before we back up
too_close_dist = height * 0.2

'how far the furthest points needs to be to move forward
far_enough_dist = height * 0.7

waiting = GetVariable("WAIT_TIMER")
delay = 250 'increase this to help avoide movement distortion making the robot go the wrong way

if (waiting = 0) then
  ' WALK AND TURN SPEED
  gBodyZTranslate = int(Interpolate(furthest_y,0,height,cMIN_BODY_Z_TRANSLATE, cMAX_BODY_Z_TRANSLATE))
  gWalk = 0
  gCrab = 0
  gTurn = 0
  gTilt = int(80) 'point the camera at the ground, you may need to reverse this
  gPan = 0
  if (dangery > too_close_dist) then
  ' nothing is danger close
    if (furthest_y > far_enough_dist) then
    ' Furthest is valid
      gWalk = int(Interpolate((furthest_y-(height * 0.5)), 0, (height * 0.5), 0, cMAX_WALK_SPEED))
      gTurn = int(Interpolate(furthest_x,0,width,-cMAX_TURN_SPEED, cMAX_TURN_SPEED))
      waiting = 1
      SetTimedVariable "WAIT_TIMER", 0, delay
      setvariable "DISPLAYTEXT", "Going for Furthest"
    else
      ' Furthest is too close (this needs to be expanded to look left/right with pan first)
      gWalk = 0
      if (left_highest_y <= right_highest_y) or (right_highest_x < (width/2)) then
        gTurn = int(Interpolate(left_highest_x,0,width,-cMAX_TURN_SPEED, cMAX_TURN_SPEED))
        setvariable "DISPLAYTEXT", "Turning Left"
      end if
      if (right_highest_y < left_highest_y) or (left_highest_x > (width/2)) then
        gTurn = int(Interpolate(right_highest_x,0,width,-cMAX_TURN_SPEED, cMAX_TURN_SPEED))
        setvariable "DISPLAYTEXT", "Turning Right"
      end if
      waiting = 1
      SetTimedVariable "WAIT_TIMER", 0, delay
    end if
  else
    'back away from danger point
    setvariable "DISPLAYTEXT", "Backing Up"
    gWalk = int(-cMAX_WALK_SPEED)
      gTurn = int(Interpolate(danger_x,0,width,-cMAX_TURN_SPEED, cMAX_TURN_SPEED))
    waiting = 1
    SetTimedVariable "WAIT_TIMER", 0, delay
  end if
'crab commands
  if (left_danger_y < height) and (right_danger_y < height) then
    if (left_danger_y < right_danger_y) then
      gCrab = int(Interpolate(left_danger_x,0,(width/2),-cMAX_CRAB_SPEED, 0))
    else
      gCrab = int(Interpolate((width - right_danger_x),0,(width/2), 0, cMAX_CRAB_SPEED))
    end if
  else
    if (left_danger_y < height) then
      gCrab = int(Interpolate(left_danger_x,0,(width/2),-cMAX_CRAB_SPEED, 0))
    end if
    if (right_danger_y < height) then
      gCrab = int(Interpolate((width - right_danger_x),0,(width/2), 0, cMAX_CRAB_SPEED))
    end If
  End if
End if

setvariable "WAIT_TIMER", waiting

'=======BINARY OBJECT AVOIDANCE END


' PIP HEAD PAN&TILT =================================================================
' Set Internal & External Variables for 'H' command (Head Pan/Tilt)
VariableH1 = (256 + int(-gPan)) and 255 
VariableH2 = (256 + int(-gTilt)) and 255 
if VariableH1 = 0 then VariableH1 = 1
if VariableH2 = 0 then VariableH2 = 1
ChkSumH = 255 - ((asc("H") + VariableH1 + VariableH2) and 255)
PIPPan = "~" & chr(3) & "H" & chr(VariableH1) & chr(VariableH2) & chr(ChkSumH)
setvariable "PIPPan", PIPPan


' Calculate the desired Body X & Z Rotate (X = Tilt, Z = Pan)
gBodyXRotate = Interpolate( gTilt, cMIN_HEAD_TILT, cMAX_HEAD_TILT, cMIN_BODY_X_ROTATE, cMAX_BODY_X_ROTATE )
gBodyZRotate = Interpolate(	gPan, cMIN_HEAD_PAN, cMAX_HEAD_PAN, cMIN_BODY_Z_ROTATE, cMAX_BODY_Z_ROTATE )


' PIP BODY ROTATE/TRANSLATE ========================================================
' Set Internal & External Variables for 'B' command (Body Pan/Tilt & Translate)
'VariableB1 = (256 + int(gBodyXRotate)) and 255
'VariableB3 = (256 + int(gBodyZRotate)) and 255
if gWalkEnable = 0 then
'	VariableB5 = (256 + int(gBodyYTranslate)) and 255
  VariableB6 = (256 + int(gBodyZTranslate)) and 255
end if
if VariableB1 = 0 then VariableB1 = 1
if VariableB3 = 0 then VariableB3 = 1
if VariableB5 = 0 then VariableB5 = 1
if VariableB6 = 0 then VariableB6 = 1
ChkSumB = 255 - ((asc("B") + VariableB1 + VariableB2 + VariableB3 + VariableB4 + VariableB5 + VariableB6) and 255)
PIPBody = "~" & chr(7) & "B" & chr(VariableB1) & chr(VariableB2) & chr(VariableB3) & chr(VariableB4) & chr(VariableB5) & chr(VariableB6) & chr(ChkSumB)
setvariable "PIPBody", PIPBody


' PIP WALK ========================================================================
' Get the WALK_ENABLE variable
gWalkEnable = GetVariable("WALK_ENABLE")

if ( gWalkEnable <> 0 ) then
'	VariableM1 = (256 + int(gCrab)) and 255
	VariableM2 = (256 + int(gWalk)) and 255
	VariableM3 = (256 + int(gTurn)) and 255
end if

if VariableM2 = 0 then VariableM2 = 1
if VariableM3 = 0 then VariableM3 = 1


setvariable "GWALK", gWalk
setvariable "GTURN", gTurn
ChkSumM = 255 - ((asc("M") + VariableM1 + VariableM2 + VariableM3) and 255)
if gWalkEnable = 0 or (VariableM1 = 1 and VariableM2 = 1 and VariableM3 = 1) then
	PIPMove = ""
else
	PIPMove = "~" & chr(4) & "M" & chr(VariableM1) & chr(VariableM2) & chr(VariableM3) & chr(ChkSumM)
end if
setvariable "PIPMove", PIPMove




' PIP WAKE COMMAND ========================================================================

gPowerOn = GetVariable("POWER_ON")
gPowerOff = GetVariable("POWER_OFF")
PIPCommand = ""
if gPowerOn then
	PIPCommand = "~" & chr(1) & "+" & chr(212)
	gPowerOn = 0
	gPowerState = 1
	setvariable "POWER_ON", gPowerOn
end if

if gPowerOff then
	PIPCommand = "~" & chr(1) & "-" & chr(210)
	gPowerOff = 0
	gPowerState = 0
	gWalkEnable = 0
	SetVariable "WALK_ENABLE", gWalkEnable
	setvariable "POWER_OFF", gPowerOff
end if

setvariable "PIPCommand", PIPCommand
setvariable "POWER_STATE", gPowerState






' =============================================================================
' FUNCTIONS
' =============================================================================

' LINEAR INTERPOLATION FUNCTION
function Interpolate( pIn, pInLow, pInHigh, pOutLow, pOutHigh )

if pIn > pInHigh then 
	Interpolate = pOutHigh 
elseif pIn < pInLow then 
	Interpolate = pOutLow 
else 	
	Interpolate = ( pOutLow + ( ( (pIn - pInLow) * (pOutHigh - pOutLow) ) / (pInHigh - pInLow) ) )
end if
	
end function


